/* Copyright (c) 2004, Jason Wolfe  <jawolfe@uclink.berkeley.edu>

   Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
   associated documentation files (the "Software"), to deal in the Software without restriction, including 
   without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
   copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the 
   following conditions:

   The above copyright notice and this permission notice shall be included in all copies or substantial 
   portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
   TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS IN THE SOFTWARE.

   ( This is the BSD/MIT Open Source Licence,
     http://www.opensource.org/licenses/mit-license.php )
 */

/*  Bootstrap.java
 *
 *  Contains static methods for reading data, creating bootstrap samples, and producing bias-corrected
 *  entropy and mutual information values.  
 *
 *  Defines the abstract Bootstrap.DataReader class, which reads in a dataset
 *  and processes individual items by calling Distribution.classifyData (which calls DataClassifier.classifyData in turn).
 */

import java.text.*;
import java.io.*;

public class Bootstrap {

    /*  Abstract class which trains an array of Distributions over some dataset.
     *
     *  Extensions of this class define the boostrap method, which should build on the following (pseudo)code to classify 
     *  each data item (i.e. sequence) found in the data set:
     *
        public void bootstrap(Distribution [] dists, int sInd) {
         for(each data_item in dataset) {
            for(int i = 0; i < dists.length; i++) {
                if (i >= sInd) {
                    int p = (int)random.nextPoisson();
                    if (p > 0) dist[i].classifyData(data_item,p);
                } else dist[i].classifyData(data_item,1);
            }
         }
       }
     *
     */
    public abstract static class DataReader {
        /*  Train an array of Distributions on this SCOP data set 
         */
        public void train(Distribution [] dists) { bootstrap(dists,Integer.MAX_VALUE); }

        
        /*  Train an array of Distributions
         *  <sInd> gives the first index in dists for which bootstrap sampling will be used.
         *  dists[0] ... dists[sInd-1] are trained normally (no bootstrapping)
         *  dists[sInd] ... dists[dists.length-1] are trained with bootstrap sampling.
         *  See above for skeleton code for this function
         */
        public abstract void bootstrap(Distribution [] dists, int sInd);

        // Mersenne Twister Random Number Generator used to generate bootstrap samples
        static final MersenneTwisterFast random = new MersenneTwisterFast(System.currentTimeMillis());
    }

    
    /* Method which trains Distributions in dists, makes nReps replicas of them, trains the replicas
     * with bootstrap sampling, and stores all the results in targetFolder.
     * Attempts to use as few passes on the DataReader as possible, by using as much memory as is available.
     */
    public static void train(Distribution [] dists, int nReps, DataReader reader, String targetFolder) {
        File fnew = new File(targetFolder);
        fnew.mkdir();

        // Estimate size memory usage for dists
        int totmem = 0;
        for(int i = 0; i < dists.length; i++) {
            totmem += dists[i].estMemUsage();
        }
       
        // estimate amount of available memory
        java.lang.Runtime rt = java.lang.Runtime.getRuntime();
        long freeMemory = rt.maxMemory() - (rt.totalMemory()-rt.freeMemory());        

        int maxRepsPer = Math.max((int) (freeMemory / (totmem*2)),1);
        
        int nIterations = (nReps / maxRepsPer) + 1;
        
        System.out.println("Training from data (" + nIterations + " iterations) ...");
        for(int it = 1; it <= nIterations; it++) {
            System.out.print(it + "...");
            int nRepsPer = (it == nIterations) ? nReps+1 - (nIterations-1) * maxRepsPer : maxRepsPer;
            Distribution [] reps = new Distribution [dists.length * nRepsPer];
            
            int firstRep = 0;
            if (it == 1) {
                firstRep = 1;
                System.arraycopy(dists,0,reps,0,dists.length);
            }
            for(int i = firstRep; i < nRepsPer; i++) {
                for(int j = 0; j < dists.length;j++) {
                    reps[i*dists.length + j] = dists[j].cloneEmpty();
                }
            }

            reader.bootstrap(reps,firstRep*dists.length);
            
            try {
                if (it == 1) {
                    for(int i = 0; i < dists.length;i++) {
                       reps[i].writeFile(targetFolder+"full_"+i+".dat");
                    }   
                }
                for(int i = firstRep; i < nRepsPer; i++) {
                    for(int j = 0; j < dists.length;j++) {
                       reps[i*dists.length+j].writeFile(targetFolder+"sample_"+j+"n"+(i+(it-1)*maxRepsPer)+".dat");
                    }
                }
            } catch (Exception e) {
                System.out.println();
                System.out.println("Bootstrap::train -- Error in file access: " + e);
                System.exit(0);
            }
            System.out.println();
        }
    }

    /* Load nDists trained, non-bootstrapped Distributions from folder
     * Must have been created by an earlier call to train(), and nDists must be less than or equal to 
     * dists.length on that previous call.
     */
    public static Distribution [] getDists(String folder, int nDists) {
        Distribution [] ret = new Distribution[nDists];
        for(int i = 0; i < nDists; i++) {
            try {
                ret[i] = Distribution.readFile(folder+"full_"+i+".dat");
            } catch (Exception e) {
                System.out.println("Bootstrap::getDists -- files missing or not accessible" + e);
                System.exit(0);
            }
        }
        return ret;
    }
    
    // Abstract class representing a function from Distributions to Distributions.  More explanation is given in 
    // ProcessContacts.java
    static abstract class DFunction {
        public abstract Distribution f(Distribution d);
    }
    
    // The identity DFunction
    static final DFunction identity = new DFunction() {
        public Distribution f(Distribution d) { return d;}
    };
    

    
    // Add parentheses to the string in s, inside any whitespace.  Used by bootstrapMutualInformation and bootstrapEntropy
    public static String paren(String s) {
        int l = s.length();
        String ret = "(" + s.trim() + ")                                                                               ";
        return ret.substring(0,l+2);        
    }
 
    // see below; uses identity as f
    public static void bootstrapMutualInformation(int dInd, int [] s1, int [] s2, int [] cond, String desc, String fname, int nReps) {
        bootstrapMutualInformation(dInd,s1,s2,cond,desc,fname,nReps,identity);
    }
    
    // see below; uses identity as f
    public static void bootstrapEntropy(int dInd, int [] set, int [] cond, String desc, String fname, int nfiles) {
        bootstrapEntropy(dInd,set,cond,desc,fname,nfiles,identity);
    }

    // see below; bootstrapping Mutual Information
    public static void bootstrapMutualInformation(int dInd, int [] s1, int [] s2, int [] cond, String desc, String fname, int nReps, DFunction f) {
        bootstrapQuantity(dInd,s1,s2,cond,desc,fname,nReps,f,false);
    }
    
    // see below; bootstrapping Entropy
    public static void bootstrapEntropy(int dInd, int [] set, int [] cond, String desc, String fname, int nReps, DFunction f) {
        bootstrapQuantity(dInd,set,null,cond,desc,fname,nReps,f,true);
    }
    
    /* bootstrap an entropy or mutual information value I(s1;s2|cond) on f(Distributions)
     f is a DFunction, which can do preprocessing on a Distribution 
     dnum is the number of the distribution in the original array 
       (when Bootstrap.train was called), which determines the name of its corresponding files
     s1 gives the set of dimensions in dist that make up the first informational element
     s2 gives the set of dimensions in dist that make up the second informational element (ignored if entropy=true)
     cond gives the set of dimensions in dist that are conditioned
     desc is a string description of the value
     nfiles gives the number of bootstrap samples to use
     */
    static void bootstrapQuantity(int dInd, int [] s1, int [] s2, int [] cond, String desc, String fname, int nReps, DFunction f, boolean entropy) {
        Distribution tmpDist = null;
        if (cond == null) cond = new int [0];
        try {
            tmpDist = f.f(Distribution.readFile(fname+"full_"+dInd+".dat"));
        } catch (Exception e) {
            System.out.println("Bootstrap::bootstrapMutualInformation -- files missing or not accessible: " + e);
            System.exit(0);
        }
     
        double origInf = entropy ? f.f(tmpDist).entropy(s1,cond) :
                                   f.f(tmpDist).mutualInformation(s1,s2,cond);
        
        NumberFormat nf = new DecimalFormat("0.#####");
        
        double sum = 0;
        double sqsum = 0;
        for(int i = 1; i <= nReps; i++) {
            try {
                tmpDist = f.f(Distribution.readFile(fname+"sample_"+dInd+"n"+i+".dat"));
            } catch (Exception e) {
                System.out.println("Bootstrap::bootstrapMutualInformation -- files missing or not accessible" + e);
                System.exit(0);
            }
            double inf = entropy ?  tmpDist.entropy(s1,cond) :
                                    tmpDist.mutualInformation(s1,s2,cond);
            sum += inf;
            sqsum += inf * inf;
       }
        sum/=nReps;
        double var = (sqsum/nReps) - sum * sum;

        String symbol = entropy ? "S" : "I";
        if (nReps > 0) System.out.println(symbol + paren(desc)+"\t"+nf.format((2*origInf-sum))+"\t"+nf.format(Math.sqrt(var))+"\t"+nf.format((sum-origInf)));
        else System.out.println("Calculated "+symbol + paren(desc)+": " + origInf + " (no bootstrap samples were provided)");
    }
    
};
