/* Copyright (c) 2004, Jason Wolfe  <jawolfe@uclink.berkeley.edu>

   Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
   associated documentation files (the "Software"), to deal in the Software without restriction, including 
   without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
   copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the 
   following conditions:

   The above copyright notice and this permission notice shall be included in all copies or substantial 
   portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
   TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS IN THE SOFTWARE.

   ( This is the BSD/MIT Open Source Licence,
     http://www.opensource.org/licenses/mit-license.php )
*/

/*  ContactDC.java
 *
 *  A simple implementation of the DataClassifier class which classifies a sequence of Residues, classifying each
 *  pairwise contact of the sequence as a data point.
 *  A pairwise contact is a pair of residues, where the residue earlier in the chain is designated L and
 *  the residue later in the chain is R.  L and R must be separated by at least <minSeqDist> other residues on the amino
 *  acid backbone, and fall within <maxPairDist> Angstroms distance from one another.
 *  This class is meant to be used only with the tmplReader DataReader implementation.
 *
 *  Defines an abstract method classify(Residue resL, Residue resR), which classifies a pairwise contact.  
 *
 *  Look at ProcessContacts to see the easiest way to use this class.  The return value from Classify must conform to the 
 *  number and size of the dimensions of the Distribution being trained, or can be null for an invalid data point.
 */

public abstract class ContactDC extends DataClassifier {
    double maxPairDist;     // Maximum distance between pairwise contacts, in Angstroms
    int minSeqDist;         // Minimum # of residues between pairwise contacts on the chain
    
    /** Creates a new instance of ContactDC */
    public ContactDC(double mPD, int mSD) {
        maxPairDist = mPD;
        minSeqDist = mSD;
    }
    
    /*  Classifies a sequence of residues (stored in data). 
     *  Calls the classify(Residue l, Residue) function for each valid contact pair (left, right) in the sequence
     */
    public void classifyData(Distribution d, Object data, int mul) {
        Residue [] r = (Residue []) data;
        // Compute # of contacts
        for(int i = 0; i < r.length;i++) {
            if (r[i]==null) continue;                        
            int nContacts=0;
            for(int j = i+1+minSeqDist; j < r.length; j++) {
                if (r[j]==null) continue;    
                if (dist(r[i].x,r[i].y,r[i].z,r[j].x,r[j].y,r[j].z) <= maxPairDist)
                    nContacts++;                
            }
            r[i].nContacts=nContacts;
        }
        
        // Classify contacts
        for(int i = 0; i < r.length;i++) {
            if (r[i]==null) continue;                        
            for(int j = i+1+minSeqDist; j < r.length; j++) {
                if (r[j]==null) continue;    
                if (dist(r[i].x,r[i].y,r[i].z,r[j].x,r[j].y,r[j].z) <= maxPairDist)
                    d.addDataPoint(classify(r[i],r[j]), mul);                
            }
        }
    }
    
    /* Abstract method to classify a pairwise contact into a specific bin in a Distribution.
     * Returns an int [] that specifies a specific bin in the Distribution, by assigning a value for each
     * of its dimensions
     */
    public abstract int [] classify(Residue resL, Residue resR);        
}
