/* Copyright (c) 2004, Jason Wolfe  <jawolfe@uclink.berkeley.edu>

   Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
   associated documentation files (the "Software"), to deal in the Software without restriction, including 
   without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
   copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the 
   following conditions:

   The above copyright notice and this permission notice shall be included in all copies or substantial 
   portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
   TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS IN THE SOFTWARE.

   ( This is the BSD/MIT Open Source Licence,
     http://www.opensource.org/licenses/mit-license.php )
*/

/*  Distribution.java
 *
 *  A class that represents a multi-dimensional array of counts.  In this package, each dimension represents a feature of a 
 *  residue, and each bin holds the number of times that a particular combination of feature values has been observed.
 *  
 *  A Distribution is more than a multi-dimensional array, because it can collect and classify data, and analyze the
 *  results.   The DataClassifier class acts as an interface between a Distribution and a DataReader; a DataClassifier 
 *  instance is stored in Distribution's member variable dc.
 *
 *  int [] dims holds the size of the array.  dims.length is the number of dimensions in the array, and dims[i] gives the
 *  number of possible values for the ith dimension.
 *
 *  int [] dist is the actual array of counts.  dist.length should always equal product(dims), the product of the sizes 
 *  of its dimensions.  Element [i][j][k] for an 3-dimensional array (IxJxK) is stored in dist[ (i * J + j) * K + k ], etc.
 *  
 *  
 */


import java.util.*;
import java.io.*;

public class Distribution implements Serializable {
/*************************************************************************************************************
 * Member Variables                                                                                          *
 *************************************************************************************************************/

    int [] dims;            // List of the size of each dimension in the distribution
    int [] dist;            // The actual array of counts.  dist.length = the product of the elements of dims
    DataClassifier dc;      // An object which maps from some form of input data to bins of this array 
    
/*************************************************************************************************************
 * Constructors and other methods to read/write/create Distributions                                         *
 *************************************************************************************************************/

    // Create a distribution with dimensions dms and classifier cl, using dst as the array of counts
    public Distribution(int [] dst, int [] dms, DataClassifier cl) {
        ASSERT(dst.length == product(dms));
        dims = dms;
        dist = dst;
        dc = cl;
    } 
   
    // Create a distribution of size dms with DataClassifier cl.  Starts from an empty array of counts
    public Distribution(int [] dms, DataClassifier cl) {
        dc = cl;
        dims = dms;
        int mul = 1;
        for(int i = 0; i < dims.length;i++) mul*=dims[i];
        dist = new int [mul];
    }

    // Return a copy of this distribution, with the same dimensions and DataClassifier, but with zeroed counts
    public Distribution cloneEmpty() {
        return new Distribution(dims, dc);
    }

    
    /* Write this distribution to a file  */
    public void writeFile(String fname)   throws IOException, ClassNotFoundException {
        FileOutputStream fs = new FileOutputStream(fname);
        ObjectOutputStream out=new ObjectOutputStream(fs);
        out.writeObject(this);
        out.close();
        fs.close();
    }
    
    /* Read a distribution from a file and return it */
    static Distribution readFile(String fname)   throws IOException, ClassNotFoundException {
        FileInputStream fin=new FileInputStream(fname);
        ObjectInputStream oin=new ObjectInputStream(fin);
        Distribution ret=(Distribution)(oin.readObject());
        oin.close();
        fin.close();
        return ret;
    }

/*************************************************************************************************************
 * Methods to add data to the Distribution                                                                   *
 *************************************************************************************************************/
    
    // Process some new data item, giving it a weight of mul 
    // This function is called from within a DataReader.  The dataReader should provide a compatible data Object.
    public void classifyData(Object data, int mul) {
        dc.classifyData(this,data,mul);
    }

    /* Add a single observation count to a specified bin
     * Array dp specifies a bin by assigning values for each of the dimensions in dims
     */
    public void addDataPoint(int [] dp) {
        addDataPoint(dp,1);
    }
    
    /* Add m observation counts to a specified bin
     * Array dp specifies a bin by assigning values for each of the dimensions in dims
     */
    public void addDataPoint(int [] dp,int m) {
        if (dp==null) return;
        ASSERT (dp.length == dims.length);
        int sum=0;
        for(int i = 0; i < dp.length;i++) {
            ASSERT(dp[i] >= 0 && dp[i] < dims[i]);
            sum = sum * dims[i] + dp[i];
        }
        dist[sum]+=m;
    }

/*************************************************************************************************************
 * Methods to get data out of the Distribution                                                               *
 *************************************************************************************************************/

    // return an element of dist by index.  
    public int get(int ind) {
        return dist[ind];
    }
     
    // Very approximate upper bound of memory used by this distribution (in bytes)
    public int estMemUsage() {
        return 4 * (dist.length + dims.length)+5000;
    }
   
    /* Compute the total number of counts in the Distribution, which represents the total number of data points
     */
    public int sumDist() {
        if (dims.length==0) return 0;
        int sum = 0;
        for(int i = 0; i < dist.length;i++) {
            sum+=dist[i];
        }
        return sum;
    }
    
    /*  Collapse the Distribution down to a single dimension, accumulating counts along all other dimensions
     */
    public int [] collapseDist(int dim) {
        int [] ret = new int [dims[dim]];
        int skip = 1;
        for(int i = dim+1; i < dims.length; i++) skip *= dims[i];
        for(int i = 0; i < dist.length; i++) {
            int newbin = (i / skip) % dims[dim];
            ret[newbin] += dist[i];
        }
        return ret;
    }

    
    /* Collapse out a single dimension of this distribution, accumulating its counts
     */
    public Distribution sumOutDim(int dim) {
        int [] ret = new int [dist.length/dims[dim]];
        //find the "skip" for the dimension being summed out, the
        //product of the dimensions smaller than it
        int skip = 1;
        for(int i = dim+1; i < dims.length; i++) skip *= dims[i];
        for(int i = 0; i < dist.length; i++) {
            //Give the proper bin in the new distribution this bin's contents
            int lowpart = i % skip;
            int highpart = i / (skip*dims[dim]);
            int newbin = lowpart + highpart * skip;
            ret[newbin] += dist[i];
        }
        return new Distribution(ret,removeArr(dims,dim),dc);
    }
    
    /*  Returns an n-1 dimensional slice of this distribution, given a value for a dimension
     */
    public Distribution subDist(int dim, int val) {
        ASSERT(val >= 0 && val < dims[dim]);
        int [] ret = new int [dist.length/dims[dim]];
        int skip = 1;
        for(int i = dim+1; i < dims.length; i++) skip *= dims[i];
        for(int i = 0; i < ret.length; i++) {
            int lowpart = i % skip;
            int highpart = i / (skip);
            int oldbin = lowpart + highpart * (skip*dims[dim]) + val*skip;
            ret[i] = dist[oldbin];
        }
        return new Distribution(ret,removeArr(dims,dim),dc);
    }
  
    /* Remaps the possible values for a dimension, possibly merging or discarding values.
       
       Newvals is an array which must be of length dims[dim], which
     specifies a function from old dimension values to new
     ones.  The set of new values must include every integer from 0 to n,
     for some n <= dims.length, and no numbers > n
     
     A value can be mapped to -1 to eliminate it altogether, discarding its counts. 
     */
    public Distribution remapDim(int dim, int [] newvals) {
        if (newvals.length != dims[dim]) throw new IllegalArgumentException("remapDim");
        boolean [] vals = new boolean[dims[dim]];
        for(int i = 0; i < newvals.length;i++) {
            if (newvals[i]==-1) continue;
            vals[newvals[i]] = true;
        }
        int cutoff=-1;
        for(int i = 0; i < vals.length;i++) {
            if (cutoff==-1) {
                if (!vals[i]) cutoff=i;
            } else {
                if (vals[i]) throw new IllegalArgumentException();
            }
        }
        if (cutoff==-1) cutoff = vals.length;
        int [] newdims = new int[dims.length];
        System.arraycopy(dims,0,newdims,0,dims.length);
        newdims[dim] = cutoff;
        Distribution newDist = new Distribution(newdims, dc);
        
        int skip = 1;
        for(int i = dim+1; i < dims.length; i++) skip *= dims[i];
        for(int i = 0; i < dist.length; i++) {
            int dimVal = (i / skip) % dims[dim];
            int newDimVal = newvals[dimVal];
            if (newDimVal==-1) continue;
            int newbin = (i / (skip*dims[dim])) * (skip*cutoff) + newDimVal*skip + (i % skip);
            newDist.dist[newbin] += dist[i];
        }
        
        return newDist;
    }
    
/*************************************************************************************************************
 * Methods to compute entropy and mutual information values                                                   *
 *************************************************************************************************************/
 
    
    /* Return the total entropy in this Distribution
     **/
    public double entropy() {
        double tot = sumDist();
        double ent = 0;
        for(int i = 0; i < dist.length; i++) {
            if (dist[i] == 0) continue;
            ent += dist[i] * Math.log(dist[i]/tot);
        }
        return -1 * ent / (tot * Math.log(2));
    }
    
    /* Return the entropy in a subset of the dimensions of a distribution
     **/
    public double entropy(int [] set) {
        set = sortedSet(set);

        // Sum out irrelevant dimensions
        int sp = set.length-1;
        Distribution d = this;
        for(int i = dims.length-1; i >= 0; i--) {
            if (sp >= 0 && set[sp] == i) {
                sp--;
            } else {
                d = d.sumOutDim(i);
            }
        }
        return d.entropy();
    }
    
    /* Return the entropy in a subset of the dimensions of a distribution, conditioned on some other 
     * subset of its dimensions 
     **/
    public double entropy(int [] set, int [] cond) {
        set = sortedSet(set);
        if (cond == null || cond.length == 0) return entropy(set);
        cond = sortedSet(cond);
        
        if (overlap(set,cond)) throw new IllegalArgumentException("Distribution:entropy(int [] set, int [] cond): set and cond overlap");

        int [] merged = merge(set,cond);
        Arrays.sort(merged);

        
        Distribution d = this;
        int mp = merged.length-1;
        for(int i = dims.length-1; i >= 0; i--) {
            if (mp >= 0 && merged[mp] == i) {
                mp--;
            } else {
                d = d.sumOutDim(i);
                cond = correctArr(cond,i);
            }
        }
       
        return d.conditionEntropy(cond.length-1, cond);
    }
    
    /* Helper function to condition entropy on the set of dimensions in cond
     **/
    public double conditionEntropy(int cind, int [] cond) {
        if (cind < 0) return entropy();
        int tot_i = 0;
        double tot_ent = 0;
        for(int i = 0; i < dims[cond[cind]]; i++) {
            Distribution tmp = subDist(cond[cind],i);
            int tmp_sd = tmp.sumDist();
            if (tmp_sd == 0) continue;
            tot_i += tmp_sd;
            tot_ent += tmp_sd * tmp.conditionEntropy(cind-1,cond);
        }
        return tot_ent / tot_i;
    }
    
    // Compute the mutual information between ests of dimensions s1 and s2
    public double mutualInformation(int [] s1, int [] s2) { return mutualInformation(s1,s2,new int [] {});}
    
    
    // Compute the mutual information between ests of dimensions s1 and s2, conditioning on set cond
    public double mutualInformation(int [] s1, int [] s2, int [] cond) {
        s1 = sortedSet(s1);
        s2 = sortedSet(s2);
        cond = sortedSet(cond);
        
        if (overlap(s1,cond)) throw new IllegalArgumentException("Distribution:mutualInformation(int [] s1, int [] s2, int [] cond): s1 and cond overlap");
        if (overlap(cond,s2)) throw new IllegalArgumentException("Distribution:mutualInformation(int [] s1, int [] s2, int [] cond): s2 and cond overlap");

        int [] merged = sortedSet(merge(s1,s2));
        return entropy(s1,cond) + entropy(s2,cond) - entropy(merged,cond);
    }
    
 
/*************************************************************************************************************
 * Utility Methods (not public)                                                                              *
 *************************************************************************************************************/    
    
    /* Debug function similar to c's assert */
    static void ASSERT(boolean b) {
        if (!b) throw new IllegalArgumentException();
    }


    /* Helper function for entropy(...) and mutualInformation(...)
     * Takes a sorted array of integers, and returns it as-is if it contains no duplicates.
     * If the sorted_arr contians duplicates, this routine creates a copy of sorted_arr without 
     * duplicates and returns it
     */
    static int [] removeDups(int [] sorted_arr) {
        if (sorted_arr.length < 2) return sorted_arr;
        int dc = 0;
        int last = sorted_arr[0];
        for(int i = 1; i < sorted_arr.length; i++) {
            if (sorted_arr[i] == last) dc++;
            else last = sorted_arr[i];
        }
        if (dc == 0) return sorted_arr;
        int [] ret = new int[sorted_arr.length-dc];
        int r_ind = 0;
        last = ret[r_ind++] = sorted_arr[0];
        for(int i = 1; i < sorted_arr.length; i++) {
            if (sorted_arr[i] != last) {
                last = ret[r_ind++] = sorted_arr[i];                 
            }
        }
        return ret;
    }
    
    /* Helper function for entropy(...) and mutualInformation(...)
     * Clone the array of ints arr, sort it, and remove duplicates
     */
    static int [] sortedSet(int [] arr) {
        int [] ret = (int []) arr.clone();
        Arrays.sort(ret);
        return removeDups(ret);
    }
    
    /* Helper function for entropy(...) and mutualInformation(...)
     * Copy the contents of two arrays to a single array (sequentially)
     */
    static int [] merge(int [] a1, int [] a2) {
        int [] ret = new int [a1.length+a2.length];
        System.arraycopy(a1,0,ret,0,a1.length);
        System.arraycopy(a2,0,ret,a1.length,a2.length);
        return ret;
    }

    /* Helper function for entropy(...) and mutualInformation(...)
     * Takes in two sorted arrays of ints, s1 and s2, and returns true iff they share any elements
     */
    static boolean overlap(int [] s1, int [] s2) {
        int i1 = 0, i2 = 0;
        while(i1 < s1.length && i2 < s2.length) {
            if (s1[i1] == s2[i2]) return true;
            if (s1[i1] < s2[i2]) i1++;
            else i2++;
        }
        return false;
    }
    
    /* Helper function for entropy(...) and mutualInformation(...)
     * Corrects an array for the removal of a dimension from the Distribution 
     * Precondition: no element of arr equals rdim 
     * Substracts 1 from all elements of arr that are greater than rdim.
     */
    static int [] correctArr(int [] arr, int rdim) {
        int [] ret = new int [arr.length];
        for(int i = 0; i < arr.length; i++) {
            if (arr[i] > rdim) ret[i] = arr[i]-1;
            else ret[i] = arr[i];
        }
        return ret;
    }    
    
    /* Helper function for entropy(...) and mutualInformation(...)
     * Removes element ind from array arr 
     */
    static int [] removeArr(int [] arr, int ind) {
        int [] ret = new int [arr.length-1];
        System.arraycopy(arr,0,ret,0,ind);
        System.arraycopy(arr,ind+1,ret,ind,ret.length-ind);
        return ret;
    }

    /* return the product of all elements of arr
     **/
    static int product(int [] arr) {
        int ret = 1;
        for(int i = 0; i < arr.length; i++) ret *= arr[i];
        return ret;
    }
}
