/* Copyright (c) 2004, Jason Wolfe  <jawolfe@uclink.berkeley.edu>

   Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
   associated documentation files (the "Software"), to deal in the Software without restriction, including 
   without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
   copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the 
   following conditions:

   The above copyright notice and this permission notice shall be included in all copies or substantial 
   portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
   TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS IN THE SOFTWARE.

   ( This is the BSD/MIT Open Source Licence,
     http://www.opensource.org/licenses/mit-license.php )
*/

/*  ProcessContacts.java
 *
 *  Essentially just a Main function, which sets up Distributions, trains them, and analyzes them.
 *  
 *  This class contains all the high-level code that calculates the values given in the paper.
 */

import java.util.*;
import java.io.*;

public class ProcessContacts {
/*************************************************************************************************************
 *  Alphabets and constants                                                                                  *
 *************************************************************************************************************/
    static final String AAs = "ACDEFGHIKLMNPQRSTVWY";       // Amino acid alphabet
    
    static final String STypes_red8  = "HGIEBTSL";          // 8-state SS alphabet
    static final String STypes_red3 = "HEL";                // 3-State reduced SS alphabet
    
    // Max exposed surface area for each amino acid
    static final double [] surfaceArea = {115,  135,  150,  190,  210,  75,  195,  175,  200,  170,  185,  160,  145,  180,  225,  115,  140,  155,  255,  230};
    
    // Burial thresholds
    static final double [] burialThresholds2 = {23.5};
    static final double [] burialThresholds4 = {1.5, 17.5, 40.0};
    static final double [] burialThresholds8 = {0.25, 3.25, 11.5, 23.5, 36.25, 54.25, 82.25};
    static final double [] burialThresholds100 = {0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75, 2.0, 2.25, 2.5, 2.75, 3.0, 3.75, 4.0, 4.25, 4.5, 4.75, 5.25, 5.5, 5.75, 6.0, 7.0, 7.25, 7.75, 8.0, 8.25, 8.75, 9.0, 9.25, 9.5, 10.0, 10.25, 10.5, 10.75, 11.0, 11.25, 11.5, 11.75, 12.0, 12.25, 13.0, 13.75, 14.0, 15.75, 16.0, 16.25, 16.75, 17.0, 17.25, 17.5, 18.0, 19.75, 20.25, 20.5, 20.75, 23.5, 23.75, 24.0, 25.5, 25.75, 26.0, 28.75, 29.75, 30.0, 30.25, 30.75, 31.25, 31.5, 36.25, 37.75, 38.0, 39.75, 40.0, 41.75, 42.25, 43.25, 43.5, 43.75, 44.0, 45.75, 46.0, 46.25, 48.5, 51.75, 52.0, 53.75, 54.0, 59.5, 66.0, 70.75, 71.0, 77.0, 82.25, 87.0, 87.25, 88.25, 91.0, 95.0, 100.0};
    
    // Constants defining contacts
    static final int minPairChainDist = 6;  // Min number of residues between contacts on the chain
    static final double maxPairDist = 8;    // Maximum distance between contacts, in Ang
    
    
    
/*************************************************************************************************************
 *  Methods                                                                                                  *
 *************************************************************************************************************/
    /*  Main method.
     *  Must receive one of two forms of arguments.  Execute 'Java ProcessContacts' for usage (or look around line 220)
     *  'create' creates the Distributions, and then uses Bootstrap to train them, create bootstrap samples, and
     *  write them to a file.  'analyze' skips this step.
     *
     *  Then the program performs analysis, primarily in the form of calls to Bootstrap.bootstrapMutualInformation(...).
     *  The Bootstrap class accesses the Distributions on disk produced by 'create' to calculate the results
     */

    public static void main(String [] args) {
        String tempFolder="";
        int nBootstraps=0;

        // The array of Distributions to be trained
        Distribution [] dists = new Distribution[23];
        
        if (args.length==4 && args[0].compareToIgnoreCase("create")==0) {
            tempFolder=args[2];
            nBootstraps  = Integer.parseInt(args[3]);

            
            // Set up Distributions
            
            // Local secondary structure
            dists[0] = new Distribution(new int [] {20,3,8},new LocalDC(maxPairDist,minPairChainDist) {
                public int [] classify(Residue r) {
                    int [] ret = new int[3];
                    ret[0] = r.aa;
                    ret[1] = r.stype_3;
                    ret[2] = r.stype_8;
                    return ret;
                }
            });
            
            // Local burial
            dists[1] = new Distribution(new int [] {20,2,4,8,100},new LocalDC(maxPairDist,minPairChainDist) {
                public int [] classify(Residue r) {
                    int [] ret = new int[5];
                    ret[0] = r.aa;
                    ret[1] = discretize(burialThresholds2,r.burial*100/ProcessContacts.surfaceArea[r.aa]);
                    ret[2] = discretize(burialThresholds4,r.burial*100/ProcessContacts.surfaceArea[r.aa]);
                    ret[3] = discretize(burialThresholds8,r.burial*100/ProcessContacts.surfaceArea[r.aa]);
                    ret[4] = discretize(burialThresholds100,r.burial*100/ProcessContacts.surfaceArea[r.aa]);
                    return ret;
                }
            });
            
            // Local SS and Burial
            dists[2] = new Distribution(new int [] {20,3,4}, new LocalDC(maxPairDist,minPairChainDist) {
                public int [] classify(Residue r) {
                    int [] ret = new int[3];
                    ret[0] = r.aa;
                    ret[1] = r.stype_3;
                    ret[2] = discretize(burialThresholds4,r.burial*100/ProcessContacts.surfaceArea[r.aa]);
                    return ret;
                }
            });
            
            
            // Local neighbors on the chain, L = -8 to +8
            // Each distribution holds a single neighbor.  A window of 9 residues is used for each distribution, holding the central
            // residue at the left or right side
            
            // Left-side neighbors, central residue held at right side
            for(int i = 0; i < 8; i++) {
                final int ind = i;
                dists[i+3] = new Distribution(new int [] {20,3,4,20,3,4}, new ExtLocalDC(maxPairDist,minPairChainDist,9) {
                    public int [] classify(Residue [] ress) {
                        int [] ret = new int[6];
                        ret[0] = ress[8].aa;
                        ret[1] = ress[8].stype_3;
                        ret[2] = discretize(burialThresholds4,ress[8].burial*100/surfaceArea[ress[8].aa]);
                        ret[3] = ress[ind].aa;
                        ret[4] = ress[ind].stype_3;
                        ret[5] = discretize(burialThresholds4,ress[ind].burial*100/surfaceArea[ress[ind].aa]);
                        return ret;
                    }
                });
            }
            
            // Right-side neighbors, central residue held at left side
            for(int i = 0; i < 8; i++) {
                final int ind = i;
                dists[i+11] = new Distribution(new int [] {20,3,4,20,3,4},new ExtLocalDC(maxPairDist,minPairChainDist,9) {
                    public int [] classify(Residue [] ress) {
                        int [] ret = new int[6];
                        ret[0] = ress[0].aa;
                        ret[1] = ress[0].stype_3;
                        ret[2] = discretize(burialThresholds4,ress[0].burial*100/surfaceArea[ress[0].aa]);
                        ret[3] = ress[ind+1].aa;
                        ret[4] = ress[ind+1].stype_3;
                        ret[5] = discretize(burialThresholds4,ress[ind+1].burial*100/surfaceArea[ress[ind+1].aa]);
                        return ret;
                    }
                });
            }
            
            // Correlations between a residue and a block of neighboring SS information
            dists[19] = new Distribution(new int [] {20,3,3,3,3,3,3,3,3,3},new ExtLocalDC(maxPairDist,minPairChainDist,9) {
                public int [] classify(Residue [] ress) {
                    int [] ret = new int[10];
                    ret[0] = ress[4].aa;
                    for(int i = 0; i < 9; i++) 
                        ret[i+1] = ress[i].stype_3;
                    return ret;
                }
            });

            // Correlations between a residue and a block of neighboring burial information
            dists[20] = new Distribution(new int [] {20,4,4,4}, new ExtLocalDC(maxPairDist,minPairChainDist,3) {
                public int [] classify(Residue [] ress) {
                    int [] ret = new int[4];
                    ret[0] = ress[1].aa;
                    for(int i = 0; i < 3; i++) 
                        ret[i+1] = discretize(burialThresholds4,ress[i].burial*100/surfaceArea[ress[i].aa]);
                    return ret;
                }
            });
            
            // Nearest-neighbor residues identities and structural information
            // This is similar to two of the above distributions, but uses a moving window of only 2 residues
            dists[21] = new Distribution(new int [] {20,3,4,20,3,4},new ExtLocalDC(maxPairDist,minPairChainDist,2) {
                public int [] classify(Residue [] ress) {
                    int [] ret = new int[6];
                    ret[0] = ress[0].aa;
                    ret[1] = ress[0].stype_3;
                    ret[2] = discretize(burialThresholds4,ress[0].burial*100/surfaceArea[ress[0].aa]);
                    ret[3] = ress[1].aa;
                    ret[4] = ress[1].stype_3;
                    ret[5] = discretize(burialThresholds4,ress[1].burial*100/surfaceArea[ress[1].aa]);
                    return ret;
                }
            });
            
            
            dists[22] = new Distribution(new int [] {20,3,4,20,3,4},
                                                      new ContactDC(maxPairDist,minPairChainDist) {
               public int [] classify(Residue resL, Residue resR) {
                    int [] ret = new int[6];
                    ret[0] = resL.aa;
                    ret[1] = resL.stype_3;
                    ret[2] = discretize(burialThresholds4,resL.burial*100/surfaceArea[resL.aa]);
                    ret[3] = resR.aa;
                    ret[4] = resR.stype_3;
                    ret[5] = discretize(burialThresholds4,resR.burial*100/surfaceArea[resR.aa]);
                    return ret;
               }
            });
           
            // Bin the dataset into the above Distributions, plus nBootstrap bootstraped copies, and 
            // save the results to tempFolder
            Bootstrap.train(dists,nBootstraps,new tmplReader(args[1]),tempFolder);

        } else if (args.length==3 && args[0].compareToIgnoreCase("analyze")==0) {
            tempFolder=args[1];
            nBootstraps = Integer.parseInt(args[2]);
        }  else {
            // Incorrect syntax
            System.out.println("Usage: ");
            System.out.println("ProcessContacts create <dataset folder> <folder for temporary files> <number of bootstrap samples>");
            System.out.println("ProcessContacts analyze <folder for temporary files> <number of bootstrap samples>");
            System.out.println();
            System.out.println("The create command bins the sequence data files, and the desired number of bootstrap samples of the sequence data files.");
            System.out.println("It then outputs the binned results to data files in the folder for temporary files, and then analyzes the results.");
            System.out.println("The analyze command loads the binned results (and boostraped results) directly from the temporary folder, and then performs the same analysis as the create command.");
            
            System.out.println("  All folder names MUST contain a trailing slash.");
            System.exit(1);
        }
        
        // Get a copy of the un-bootstrapped distributions from disk
        dists = Bootstrap.getDists(tempFolder, dists.length);

        // Do analysis
        
        System.out.println("# All secondary structure uses the 3-state FA alphabet and 4 burial bins unless noted otherwise");
        int numR = dists[0].sumDist();
        System.out.println("# Total # of residues: " + numR);
        int numC = dists[22].sumDist();
        System.out.println("# Total # of contacts within " + maxPairDist + " Angstroms, separated by at least " + minPairChainDist + " other residues on the chain: " + numC);
        System.out.println("# Contacts per residue: " + (numC * 1. / numR));
        System.out.println();
        System.out.println();
        
        System.out.println("# Counts from dataset");
        System.out.println();

        System.out.println("# Amino Acid Counts");
        int [] resCnts = dists[0].collapseDist(0);
        for(int i = 0; i < 20; i++) System.out.println(AAs.charAt(i) + "\t" + resCnts[i]);
        System.out.println();
        System.out.println();
        
        System.out.println("# Amino Acid X Secondary Structure Counts");
        System.out.print("#AA");
        for(int i = 0; i < STypes_red3.length(); i++) System.out.print("\t" + STypes_red3.charAt(i));
        System.out.println();
        Distribution SSDist = dists[2].sumOutDim(2);
        for(int i = 0; i < 20; i++) {
            int [] SSCnts = SSDist.subDist(0,i).collapseDist(0);
            System.out.print(AAs.charAt(i));
            for(int j = 0; j < SSCnts.length; j++) System.out.print("\t" + SSCnts[j]);
            System.out.println();
        }
        System.out.println();
        System.out.println();

        System.out.println("# Amino Acid X Burial Counts (4 bins)");
        System.out.print("#AA");
        for(int i = 1; i <= 4; i++) System.out.print("\tBin " + i);
        System.out.println();
        Distribution BurialDist = dists[2].sumOutDim(1);
        for(int i = 0; i < 20; i++) {
            int [] BurialCnts = BurialDist.subDist(0,i).collapseDist(0);
            System.out.print(AAs.charAt(i));
            for(int j = 0; j < BurialCnts.length; j++) System.out.print("\t" + BurialCnts[j]);
            System.out.println();
        }
        System.out.println();
        System.out.println();
        
        System.out.println("# Amino Acid X Amino Acid Counts (for contacts)");
        System.out.println("# Rows are for the identity of the earlier residue on the chain, columns the identity of the later residue");
        System.out.print("#L AA");
        for(int i = 0; i < 20; i++) System.out.print("\t" + AAs.charAt(i));
        System.out.println();
        Distribution ContactDist = dists[22].sumOutDim(5).sumOutDim(4).sumOutDim(2).sumOutDim(1);
        for(int i = 0; i < 20; i++) {
            int [] ContactCnts = ContactDist.subDist(0,i).collapseDist(0);
            System.out.print(AAs.charAt(i));
            for(int j = 0; j < ContactCnts.length; j++) System.out.print("\t" + ContactCnts[j]);
            System.out.println();
        }
        System.out.println();
        System.out.println();
        System.out.println();
        System.out.println();

        System.out.println("# All the following values are bootstrapped mutual information values.  Each line has the following format:");
        System.out.println("# <Desc> <Corrected> <S.D.> <Bias>");
        System.out.println("# <Desc> is a description of the value being calculated.");
        System.out.println("# <Corrected> is the bias-corrected M.I. value, equal to the calculated M.I. minus the observed bias.");
        System.out.println("# <S.D.> is the standard deviation of the mutual information value over the bootstrap samples.");
        System.out.println("# <Bias> is the observed bias, equal to the average M.I. value for the bootstraps minus the calculated M.I. value.");
        
        System.out.println();
        System.out.println();
        System.out.println("# Local Information");
        System.out.println();

        System.out.println("# I(R;S) for 3 versus 8 secondary structure states");
        Bootstrap.bootstrapMutualInformation(0,new int [] {0}, new int [] {1},null,"R;S(3)",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(0,new int [] {0}, new int [] {2},null,"R;S(8)",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();

        System.out.println("# I(R;B) for 2, 4, 8, and 100 burial bins (using optimal thresholds)");
        Bootstrap.bootstrapMutualInformation(1,new int [] {0}, new int [] {1},null,"R;B(2)  ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(1,new int [] {0}, new int [] {2},null,"R;B(4)  ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(1,new int [] {0}, new int [] {3},null,"R;B(8)  ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(1,new int [] {0}, new int [] {4},null,"R;B(100)",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();
        
        System.out.println("# From this point forward, all numbers use the 3-state FA Secondary Structure alphabet and 4 burial bins");
        System.out.println("# Mutual Information between R and local structure elements");
        Bootstrap.bootstrapMutualInformation(2,new int [] {0}, new int [] {1},null,  "R;S   ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(2,new int [] {0}, new int [] {2},null,  "R;B   ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(2,new int [] {0}, new int [] {1,2},null,"R;S,B ",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();
        System.out.println();
        System.out.println();

        System.out.println("# Correlations between individual neighbors L residues away on the chain and a 'central' residue");
        System.out.println();
        OutputSeqs(dists,3,0,1,"# L     I(R_i;S_[i+L])",tempFolder,nBootstraps);
        OutputSeqs(dists,3,0,2,"# L     I(R_i;B_[i+L])",tempFolder,nBootstraps);
        OutputSeqs(dists,3,1,2,"# L     I(S_i;B_[i+L])",tempFolder,nBootstraps);
 
        System.out.println();
        
        OutputSeqs(dists,3,0,0,"# L     I(R_i;R_[i+L])",tempFolder,nBootstraps);
        OutputSeqs(dists,3,1,1,"# L     I(S_i;S_[i+L])",tempFolder,nBootstraps);
        OutputSeqs(dists,3,2,2,"# L     I(B_i;B_[i+L])",tempFolder,nBootstraps);
        
        System.out.println();
        System.out.println();

        System.out.println("# Correlations between a single 'central' residue identity and blocks of neighboring structure information");
        System.out.println();
        System.out.println("# Residue versus blocks of secondary structure information");
        Bootstrap.bootstrapMutualInformation(19,new int [] {0}, new int [] {5},null,"R_i;S_i                ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(19,new int [] {0}, new int [] {4,5,6},null,"R_i;S_[i-1] ... S_[i+1]",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(19,new int [] {0}, new int [] {3,4,5,6,7},null,"R_i;S_[i-2] ... S_[i+2]",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(19,new int [] {0}, new int [] {2,3,4,5,6,7,8},null,"R_i;S_[i-3] ... S_[i+3]",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(19,new int [] {0}, new int [] {1,2,3,4,5,6,7,8,9},null,"R_i;S_[i-4] ... S_[i+4]",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();

        System.out.println("# Residue versus blocks of burial information");
        Bootstrap.bootstrapMutualInformation(20,new int [] {0}, new int [] {2},null,    "R_i;B_i                ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(20,new int [] {0}, new int [] {1,2,3},null,"R_i;B_[i-1] ... B_[i+1]",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();
        System.out.println();
        System.out.println();

        System.out.println("# Correlations between nearest-neighbors on the chain");
        System.out.println("# Features corresponding to the left residue are stated normally (R, S, B)");
        System.out.println("# Features corresponding to the right residue are stated with a prime (i.e. R', S', B')");
        System.out.println();
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {1},null,                "R;S             ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {2},null,                "R;B             ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {1,2},null,              "R;S,B           ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {4},null,                "R;S'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {5},null,                "R;B'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {4,5},null,              "R;S',B'         ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {1,4},null,              "R;S,S'          ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {2,5},null,              "R;B,B'          ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {1,2,4},null,            "R;S,B,S'        ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {1,2,5},null,            "R;S,B,B'        ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {1,2,4,5},null,          "R;S,B,S',B'     ",tempFolder,nBootstraps);
        System.out.println();
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {4},null,                "R';S'           ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {5},null,                "R';B'           ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {4,5},null,              "R';S',B'        ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {1},null,                "R';S            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {2},null,                "R';B            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {1,2},null,              "R';S,B          ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {1,4},null,              "R';S,S'         ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {2,5},null,              "R';B,B'         ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {1,4,5},null,            "R';S,S',B'      ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {2,4,5},null,            "R';B,S',B'      ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {3}, new int [] {1,2,4,5},null,          "R';S,B,S',B'    ",tempFolder,nBootstraps);
        System.out.println();
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {3},null,                "R;R'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {1}, new int [] {4},null,                "S;S'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {2}, new int [] {5},null,                "B;B'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {1}, new int [] {5},null,                "S;B'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {2}, new int [] {4},null,                "B;S'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {3},new int [] {1,4},    "R;R'|S,S'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {3},new int [] {2,5},    "R;R'|B,B'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0}, new int [] {3},new int [] {1,2,4,5},"R;R'|S,B,S',B'  ",tempFolder,nBootstraps);
        System.out.println();
        Bootstrap.bootstrapMutualInformation(21,new int [] {0,3}, new int [] {1,4},null,            "R,R';S,S'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0,3}, new int [] {2,5},null,            "R,R';B,B'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(21,new int [] {0,3}, new int [] {1,2,4,5},null,        "R,R';S,B,S',B'  ",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();
        System.out.println();
        System.out.println();
        
        System.out.println("# Inter-residue contacts separated by at least " + minPairChainDist + " residues on the chain and");
        System.out.println("# " + maxPairDist + " Angstroms in space");
        System.out.println("# Features corresponding to the residue occuring earlier in the sequence are stated normally (i.e. R, S, B)");
        System.out.println("# Features corresponding to the residue occuring later are stated with a prime (i.e. R', S', B')");
        System.out.println("# Because each contact in the dataset is a datapoint, vaules for indivdual residues are weighted by their number of contacts");
        System.out.println();
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {1},null,                "R;S             ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {2},null,                "R;B             ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {1,2},null,              "R;S,B           ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {4},null,                "R;S'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {5},null,                "R;B'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {4,5},null,              "R;S',B'         ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {1,4},null,              "R;S,S'          ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {2,5},null,              "R;B,B'          ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {1,2,4},null,            "R;S,B,S'        ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {1,2,5},null,            "R;S,B,B'        ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {1,2,4,5},null,          "R;S,B,S',B'     ",tempFolder,nBootstraps);
        System.out.println();
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {4},null,                "R';S'           ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {5},null,                "R';B'           ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {4,5},null,              "R';S',B'        ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {1},null,                "R';S            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {2},null,                "R';B            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {1,2},null,              "R';S,B          ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {1,4},null,              "R';S,S'         ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {2,5},null,              "R';B,B'         ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {1,4,5},null,            "R';S,S',B'      ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {2,4,5},null,            "R';B,S',B'      ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {3}, new int [] {1,2,4,5},null,          "R';S,B,S',B'    ",tempFolder,nBootstraps);
        System.out.println();
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {3},null,                "R;R'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {1}, new int [] {4},null,                "S;S'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {2}, new int [] {5},null,                "B;B'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {1}, new int [] {5},null,                "S;B'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {2}, new int [] {4},null,                "B;S'            ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {3},new int [] {1,4},    "R;R'|S,S'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {3},new int [] {2,5},    "R;R'|B,B'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {3},new int [] {1,2,4,5},"R;R'|S,B,S',B'  ",tempFolder,nBootstraps);
        System.out.println();
        Bootstrap.bootstrapMutualInformation(22,new int [] {0,3}, new int [] {1,4},null,            "R,R';S,S'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0,3}, new int [] {2,5},null,            "R,R';B,B'       ",tempFolder,nBootstraps);
        Bootstrap.bootstrapMutualInformation(22,new int [] {0,3}, new int [] {1,2,4,5},null,        "R,R';S,B,S',B'  ",tempFolder,nBootstraps);
        System.out.println();
        System.out.println();
        System.out.println();
        System.out.println();
        
/*      One useful feature of the code was not illustrated above.  
 *
 *      Sometimes it is desirable to do some preprocessing on the binned data, rather than rebinning it.  This can also
 *      limit the number of Distributions that have to be used for a given set of calculations.
 *
 *      This can be accomplished by extending the "Bootstrap.DFunction" class, which has a single method f.
 *      This method takes a Distribution as input, and returns some function of this Distribution as output.
 *      
 *      For example, the following DFunction would remove all contacts involving a residue with SS type of "L" from
 *      Distribution 22 above (the ContactDC)
 
        class RemoveL1 extends Bootstrap.DFunction {
            public Distribution f (Distribution x) {
                return x.remapDim(1,new int [] {0,1,-1}).remapDim(4,new int [] {0,1,-1});
            }
        }
 
 *      DFunctions can be passed to bootstrapMutualInformation or bootstrapEntropy, and are applied (to 
 *      the Distribution and bootstrap sample Distributions) before the mutual information or entropy calculation.     
 *
 *      For example, the following command would measure the mutual information between residue identities, 
 *      for contacts between residues with secondary structure type "H" or "E"
 
        Bootstrap.bootstrapMutualInformation(22,new int [] {0}, new int [] {3},null,"R1 ; R2 | S1!=L AND S2 !=L",tempFolder,nBootstraps,new RemoveL1());

 */
    }
    

    
    /*  Helper function for calculating I(Vi1_i;Vi2_[i+L]) for L=-8 thru +8 on the "local neighbors" above.
     */
    public static void OutputSeqs(Distribution [] dists, int distInd, int i1, int i2, String label, String tempFolder, int nBootstraps) {
        System.out.println(label);

        // Left side
        for(int i = 0; i < 8; i++) {
            Bootstrap.bootstrapMutualInformation(distInd + i,new int [] {i1}, new int [] {i2+3}, null, ""+(i-8),tempFolder,nBootstraps);
        }
        
        // Center
        if (i1 != i2) Bootstrap.bootstrapMutualInformation(distInd + 7,new int [] {i1}, new int [] {i2}, null, "0",tempFolder,nBootstraps);
        else Bootstrap.bootstrapEntropy(distInd + 7,new int [] {i1}, null, "0",tempFolder,nBootstraps);
        
        // Right side
        for(int i = 0; i < 8; i++) {
            Bootstrap.bootstrapMutualInformation(distInd + i+8,new int [] {i1}, new int [] {i2+3}, null, ""+(i+1),tempFolder,nBootstraps);
        }
        
        System.out.println();
        System.out.println();
    }
    
    
    
};